#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gnome.h>

/* Don't enable this unless you know what you're doing :) */
//#define MEMORY_DEBUG

typedef struct {
   void *mem;
   unsigned int size;
   const char *file;
   int line;
} MemoryChunk;

unsigned int total = 0;

GList *mem_list = NULL;

#define DEBUG __FILE__,__LINE__,

#ifdef MEMORY_DEBUG
#define d_assert(exp) if (!(exp)) { fprintf(stderr, "%s, %s:%i: assertion failed: %s\n", __PRETTY_FUNCTION__,file,line,#exp); return NULL; }
#else
#define d_assert(exp) if (!(exp)) { return NULL; }
#endif

void d_warning(const char *file, int line, char *func, char *msg, void *mem) {
   fprintf(stderr, "%s(0x%08x, %s:%i): %s\n", func, (unsigned int)mem, 
	   file, line, msg);
}

void *d_memory_hunk_append(const char *file, int line, void *ptr, unsigned int size) {
#ifdef MEMORY_DEBUG
   MemoryChunk *chunk;
   
   d_assert(ptr);
   d_assert(size > 0);
   
   chunk = malloc(sizeof(MemoryChunk));
   d_assert(chunk);
   
   chunk->mem = ptr;
   chunk->size = size;
   chunk->file = file;
   chunk->line = line;
   
   total += chunk->size;
   
   mem_list = g_list_append(mem_list, chunk);
#endif
   
   return NULL;
}

char *d_config_get_string(const char *file, int line, const char *key) {
   char *ret;

   ret = gnome_config_get_string(key);
   d_assert(ret);

   d_memory_hunk_append(file, line, ret, strlen(ret) + 1);
   
   return ret;
}

void *d_hunk_add(const char *file, int line, void *ptr, unsigned int size) {
   d_memory_hunk_append(file, line, ptr, size);
   
   return NULL;
}

void *d_malloc(const char *file, int line, unsigned int size) {
   void *mem;
   
   d_assert(size > 0);
   
   mem = malloc(size);
   d_assert(mem);

   d_memory_hunk_append(file, line, mem, size);
   
   return mem;
}

void *d_calloc(const char *file, int line, unsigned int size) {
   void *mem;

   d_assert(size > 0);
   
   mem = calloc(1, size);
   d_assert(mem);

   d_memory_hunk_append(file, line, mem, size);
   
   return mem;
}

char *d_strdup_vprintf(const char *file, int line, const char *s, va_list args) {
   char *mem;

   d_assert(s);
   
   mem = g_strdup_vprintf(s, args);
   d_assert(mem);
   
   d_memory_hunk_append(file, line, mem, strlen(mem) + 1);
   
   return mem;
}

char *d_strdup_printf(const char *file, int line, const char *s, ...) {
   va_list args;
   char *mem;

   d_assert(s);
   
   va_start(args, s);
   mem = g_strdup_vprintf(s, args);
   va_end(args);
   d_assert(mem);
   
   d_memory_hunk_append(file, line, mem, strlen(mem) + 1);
   
   return mem;
}

MemoryChunk *d_find_chunk(void *mem) {
   MemoryChunk *chunk;
   GList *ptr;
   
   for(ptr=mem_list; ptr; ptr=ptr->next) {
      chunk = ptr->data;
      if (!chunk)
	continue;
      
      if (chunk->mem == mem)
	return chunk;
   }
   
   return NULL;
}

void *d_free(const char *file, int line, void *ptr) {
   MemoryChunk *chunk;
   GList *link;

   d_assert(ptr);
   
#ifdef MEMORY_DEBUG
   chunk = d_find_chunk(ptr);
   if (!chunk) {
      d_warning(file, line, "d_free", "attempting to free memory not allocated by a debug function",
		ptr);
   }
   
   link = g_list_find(mem_list, chunk);
   mem_list = g_list_remove_link(mem_list, link);
 
   free(ptr);
   free(chunk);
#else
   /* keep the compiler happy */
   chunk = NULL;
   link = NULL;
   
   free(ptr);
#endif
   
   return NULL;
}

char *d_strdup(const char *file, int line, const char *s) {
   char *mem;

   d_assert(s);
   
   mem = strdup(s);
   d_assert(mem);
   
   d_memory_hunk_append(file, line, mem, strlen(mem) + 1);
   
   return mem;
}

void d_summary_calc_data(MemoryChunk *chunk, int *r, int *a, unsigned int *s) {
   GList *ptr;
   MemoryChunk *c;
   int repeat = 0, avg;
   unsigned int size = 0;
   
   for(ptr=mem_list; ptr; ptr=ptr->next) {
      c = ptr->data;
      if (!strcmp(chunk->file, c->file) &&
	  chunk->line == c->line) {
	 repeat++;
	 size += chunk->size;
      }
   }
   
   avg = size / repeat;
   
   *r = repeat;
   *s = size;
   *a = avg;
}

void d_summary() {
#ifdef MEMORY_DEBUG
   MemoryChunk *chunk, *pchunk;
   GList *ptr, *pptr, *printed = NULL;
   unsigned int total_unfreed = 0;
   int already_printed;
   
   printf("Non-freed stats\n");
   printf("Address   \tFile         Line\tRepeat\tAvg  \tTotal Size\n");
   for(ptr=mem_list; ptr; ptr=ptr->next) {
      chunk = ptr->data;
      if (!chunk)
	continue;

      total_unfreed += chunk->size;
      for(pptr=printed, already_printed=0; pptr; pptr=pptr->next) {
	 pchunk = pptr->data;
	 if (!strcmp(chunk->file, pchunk->file) &&
	     chunk->line == pchunk->line)
	   already_printed = 1;
      }
      if (!already_printed) {
	 int repeat = -1, avg = -1;
	 unsigned int size = -1;
	 
	 d_summary_calc_data(chunk, &repeat, &avg, &size);
	 printf("0x%08x\t%-12s %i\t%i\t%i\t%i\n", (unsigned int)chunk->mem,
		chunk->file, chunk->line, repeat, avg, size);
	 printed = g_list_append(printed, chunk);
      }
   }
   
   printf("total allocated: %i\n", total);
   printf("total not freed: %i\n", total_unfreed);
#endif
}
